#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <string.h>
#include "array_utils.h"

#define MIN_SIZE 7
#define MIN_RATIO 0.9
#define MAX_RATIO 1.1
#define VERB 0

void insertion_sort(int a[],int left,int right)
{	
	if(left >=right)
		return;

	int i;
	//1. inserisci tutti gli interi da left+1..right
	for(i=left+1;i <=right;i++)
	{
		//2. trova la posizione in cui mettere a[i]
		int curr_val=a[i];
		int j=i-1;
		while(j >=left && a[j] >curr_val)
		{
			a[j+1] =a[j];
			j--;
		}
		
		//la posizione in cui mettere a[i] e' sempre a[j+1] !!
		a[j+1]=curr_val;
	}
}

//Fonde a[left..mid] con a[mid+1..right]
void merge(int a[],int left,int mid,int right)
{
	if(VERB){
		printf("merge(): begin left=%d mid=%d right=%d\n",left,mid,right);
		print_array2(a,left,right);
	}
	
	//le lunghezze della meta' sinistra dell'array e della meta' destra
	int len1 =mid-left+1;
	int len2 =right-mid; //right-(mid+1)+1

	//1. copia a[left..mid] in un arry temporaneo l
	int* l =malloc(len1 * sizeof(int));
	int i;
	for(i=0;i <len1;i++)
		l[i]=a[left+i];
		
	if(VERB){
		printf("merge(): len1=%d\n",len1);
		print_array(l,len1);
	}
		
	//2. copia a[mid+1..right] in un array temporaneo r
	int* r=malloc(len2 * sizeof(int));
	for(i=0;i <len2;i++)
		r[i]=a[mid+1+i];
		
	if(VERB){
		printf("merge(): len2=%d\n",len2);
		print_array(r,len2);
	}
	
	//3. fondi gli array temporanei e memorizza la sequenza ordinata in a[left..right]
	i=0; //indice del prossimo elemento di l da copiare
	int j=0; //indice del prossimo elemento di r da copiare
	int k=left; //indice del prossimo elemento che verra' scritto in a[left..right] 
	while(i <len1 && j <len2)
	{
		if(l[i] <=r[j])
		{
			a[k]=l[i];
			i++;
		}
		else
		{
			a[k]=r[j];
			j++;
		}
		
		k++;
	}
	
	//4. copia i rimanenti elementi in l[i..len1-1] (potremmo usare memcpy()!! )
	for(;i <len1;i++)
	{
		a[k]=l[i];
		k++;
	}

	//5. copia i rimanenti elementi in r[j..len2-1]
	for(;j <len2;j++)
	{
		a[k]=r[j];
		k++;
	}
	
	if(VERB){
		printf("merge(): end\n");
		print_array2(a,left,right);
	}
	
	//4. libera lo spazio necessario per gli array temporanei
	free(l);
	free(r);
}

void merge_sort(int a[],int left,int right)
{
	if(VERB){
		printf("merge_sort(): begin left=%d right=%d\n",left,right);
		print_array2(a,left,right);
	}

	//caso base
	if(left >=right)
	{
		if(VERB) printf("merge_sort(): end (left>=right)\n");
		return;
	}
	
	//OPTIMIZATION
	if(right-left+1 <=MIN_SIZE)
	{
		insertion_sort(a,left,right);
		if(VERB) print_array2(a,left,right);
		return;
	}
	
	//recursive call
	if(VERB) printf("merge_sort(): recursive call\n");
	
	int mid=(left+right)/2;
	//1. sort the first half
	merge_sort(a,left,mid);
	
	//2. sort the snd half
	merge_sort(a,mid+1,right);
	
	//3. merge the two halves
	merge(a,left,mid,right);
	
	if(VERB){
		printf("merge_sort(): end left=%d right=%d\n",left,right);
		print_array2(a,left,right);
	}
}

/**
* Assume che right >left. (i.e. che a[left..right] contenga almeno 2 elementi)
* Sceglie a[pivot_idx] come pivot.
* Partiziona ali elementi in a[left..right] mettendo in 
* a[left..i] gli elementi di valore minore o uguale al pivot, ed in 
* a[i+1..right] gli elementi strettamenti maggiori del pivot.
* Restituisce l'indice i.
*/
int partition(int a[],int left,int right,int pivot_idx)
{
	if(VERB){
		printf("partition(): begin left=%d right=%d\n",left,right);
		print_array2(a,left,right);
	}

	//seleziona il pivot
	int tmp;
	if(pivot_idx !=right)
	{
		tmp=a[pivot_idx];
		a[pivot_idx]=a[right];
		a[right]=tmp;
	}
	
	int pivot=a[right];
	int i=left;
	int j=right-1;
	
	//a[left..i-1] contiene gli elementi <= pivot
	//a[i..j] contiene gli elementi ancora da partizionare
	//a[j+1..right] contiene gli elementi > pivot
	while(i <=j)
	{
		while(i<=j && a[i] <=pivot)
			i++;
			
		while(i<=j && a[j] >pivot)
			j--;
			
		if(i <j)
		{
			//scambia gli elementi fuori posto
			tmp=a[i];
			a[i]=a[j];
			a[j]=tmp;
		}	
	}
	
	//scambia il pivot con a[i]
	tmp =a[i];
	a[i]=pivot;
	a[right]=tmp;
	
	if(VERB){
		printf("partition(): end left=%d right=%d ret=%d\n",left,right,i);
		print_array2(a,left,right);
	}
	
	return i;
}

void quicksort(int a[],int left,int right)
{
	if(VERB){
		printf("quicksort(): begin left=%d right=%d\n",left,right);
		print_array2(a,left,right);
	}

	/*caso base*/
	if(left >=right)
	{
		if(VERB) printf("quicksort(): end (left >=right)\n");
		return;
	}
	
	/*1. partition*/
	int pivot_idx=partition(a,left,right,right/*pivot_idx*/);
	
	//OPTIMIZATION
	int len1=pivot_idx-left+1;
	int len2=right-pivot_idx; //right -(pivot_idx+1) +1
	double ratio=((double)len1) / ((double)len2);
	if(ratio < MIN_RATIO || ratio >MAX_RATIO)
	{
		merge_sort(a,left,right);
		return;
	}
	
	
	/*2. recursive calls*/
	quicksort(a,left,pivot_idx-1);
	quicksort(a,pivot_idx+1,right);
	
	if(VERB){
		printf("quicksort(): end left=%d right=%d\n",left,right);
		print_array2(a,left,right);
	}
}

int main()
{
	/*1. inizializza l'array a*/
	int seed=time(NULL);
	int len=20; //70000
	int min_v=1;
	
	//array in input NON ordinato
	//int max_v=10000;
	//int* a=random_array(len,min_v,max_v,seed);
	
	//array in input ordinato
	int delta=2;
	int* a=random_sorted_array(len,min_v,delta,seed);
	
	print_array(a,len);
	
	/*2. crea una copia di a*/
	int* a2=calloc(len,sizeof(int));
	memcpy(a2,a,sizeof(int)*len);
	assert(equals(a,a2,0,len));
	
	/*3. ordina usando MergeSort*/
	long t1=time(NULL);
	merge_sort(a,0,len-1);
	long t2=time(NULL);
	printf("MergeSort: arrayLen=%d time=%ld secs\n",len,(t2-t1));
	
	/*4. ordina usando QuickSort*/
	t1=time(NULL);
	quicksort(a2,0,len-1);
	t2=time(NULL);
	printf("QuickSort: arrayLen=%d time=%ld secs\n",len,(t2-t1));
	
	assert(equals(a,a2,0,len));
	print_array(a,len);
	print_array(a2,len);
	
	return 0;
}
